@@ -17,8 +17,8 @@
  * GNU Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this program; if not, see <https://www.gnu.org/licenses/>.
- * SPDX-License-Identifier: LGPL-2.1-or-later
+ * License along with this program; if not, write to the Free Software
+ * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
  ********************************************************************
  *
  * This code is a "clean room" implementation, written from the paper
@@ -46,7 +46,7 @@
 #include "cipher.h"
 #include "bufhelp.h"
 #include "cipher-internal.h"
-#include "bulkhelp.h"
+#include "cipher-selftest.h"
 
 
 #define TWOFISH_BLOCKSIZE 16
@@ -66,73 +66,19 @@
 #  define USE_ARM_ASM 1
 # endif
 #endif
-# if defined(__AARCH64EL__)
-#  ifdef HAVE_COMPATIBLE_GCC_AARCH64_PLATFORM_AS
-#   define USE_ARM_ASM 1
-#  endif
-# endif
-
-/* USE_AVX2 indicates whether to compile with AMD64 AVX2 code. */
-#undef USE_AVX2
-#if defined(__x86_64__) && (defined(HAVE_COMPATIBLE_GCC_AMD64_PLATFORM_AS) || \
-    defined(HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS))
-# if defined(ENABLE_AVX2_SUPPORT)
-#  define USE_AVX2 1
-# endif
-#endif
 
 
 /* Prototype for the self-test function. */
 static const char *selftest(void);
 
-
-/* Prototypes for the bulk functions. */
-static void _gcry_twofish_ctr_enc (void *context, unsigned char *ctr,
-				   void *outbuf_arg, const void *inbuf_arg,
-				   size_t nblocks);
-static void _gcry_twofish_cbc_dec (void *context, unsigned char *iv,
-				   void *outbuf_arg, const void *inbuf_arg,
-				   size_t nblocks);
-static void _gcry_twofish_cfb_dec (void *context, unsigned char *iv,
-				   void *outbuf_arg, const void *inbuf_arg,
-				   size_t nblocks);
-static size_t _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
-				       const void *inbuf_arg, size_t nblocks,
-				       int encrypt);
-static size_t _gcry_twofish_ocb_auth (gcry_cipher_hd_t c, const void *abuf_arg,
-				      size_t nblocks);
-static void _gcry_twofish_xts_crypt (void *context, unsigned char *tweak,
-				     void *outbuf_arg, const void *inbuf_arg,
-				     size_t nblocks, int encrypt);
-static void _gcry_twofish_ecb_crypt (void *context, void *outbuf_arg,
-				     const void *inbuf_arg, size_t nblocks,
-				     int encrypt);
-
 /* Structure for an expanded Twofish key.  s contains the key-dependent
  * S-boxes composed with the MDS matrix; w contains the eight "whitening"
  * subkeys, K[0] through K[7].	k holds the remaining, "round" subkeys.  Note
  * that k[i] corresponds to what the Twofish paper calls K[i+8]. */
 typedef struct {
    u32 s[4][256], w[8], k[32];
-
-#ifdef USE_AVX2
-  int use_avx2;
-#endif
 } TWOFISH_context;
 
-
-/* Assembly implementations use SystemV ABI, ABI conversion and additional
- * stack to store XMM6-XMM15 needed on Win64. */
-#undef ASM_FUNC_ABI
-#if defined(USE_AVX2)
-# ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
-#  define ASM_FUNC_ABI __attribute__((sysv_abi))
-# else
-#  define ASM_FUNC_ABI
-# endif
-#endif
-
-
 /* These two tables are the q0 and q1 permutations, exactly as described in
  * the Twofish paper. */
 
@@ -757,79 +703,15 @@ do_twofish_setkey (TWOFISH_context *ctx, const byte *key, const unsigned keylen)
 }
 
 static gcry_err_code_t
-twofish_setkey (void *context, const byte *key, unsigned int keylen,
-                cipher_bulk_ops_t *bulk_ops)
+twofish_setkey (void *context, const byte *key, unsigned int keylen)
 {
   TWOFISH_context *ctx = context;
-  unsigned int hwfeatures = _gcry_get_hw_features ();
-  int rc;
-
-  rc = do_twofish_setkey (ctx, key, keylen);
-
-#ifdef USE_AVX2
-  ctx->use_avx2 = (hwfeatures & HWF_INTEL_AVX2) != 0;
-#endif
-
-  /* Setup bulk encryption routines.  */
-  memset (bulk_ops, 0, sizeof(*bulk_ops));
-  bulk_ops->cbc_dec = _gcry_twofish_cbc_dec;
-  bulk_ops->cfb_dec = _gcry_twofish_cfb_dec;
-  bulk_ops->ctr_enc = _gcry_twofish_ctr_enc;
-  bulk_ops->ocb_crypt = _gcry_twofish_ocb_crypt;
-  bulk_ops->ocb_auth = _gcry_twofish_ocb_auth;
-  bulk_ops->xts_crypt = _gcry_twofish_xts_crypt;
-  bulk_ops->ecb_crypt = _gcry_twofish_ecb_crypt;
-
-  (void)hwfeatures;
-
+  int rc = do_twofish_setkey (ctx, key, keylen);
   _gcry_burn_stack (23+6*sizeof(void*));
   return rc;
 }
 
 
-#ifdef USE_AVX2
-/* Assembler implementations of Twofish using AVX2.  Process 16 block in
-   parallel.
- */
-extern void _gcry_twofish_avx2_blk16 (const TWOFISH_context *c, byte *out,
-				      const byte *in, int encrypt) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_ctr_enc(const TWOFISH_context *ctx,
-				       unsigned char *out,
-				       const unsigned char *in,
-				       unsigned char *ctr) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_cbc_dec(const TWOFISH_context *ctx,
-				       unsigned char *out,
-				       const unsigned char *in,
-				       unsigned char *iv) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_cfb_dec(const TWOFISH_context *ctx,
-				       unsigned char *out,
-				       const unsigned char *in,
-				       unsigned char *iv) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_ocb_enc(const TWOFISH_context *ctx,
-				       unsigned char *out,
-				       const unsigned char *in,
-				       unsigned char *offset,
-				       unsigned char *checksum,
-				       const u64 Ls[16]) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_ocb_dec(const TWOFISH_context *ctx,
-				       unsigned char *out,
-				       const unsigned char *in,
-				       unsigned char *offset,
-				       unsigned char *checksum,
-				       const u64 Ls[16]) ASM_FUNC_ABI;
-
-extern void _gcry_twofish_avx2_ocb_auth(const TWOFISH_context *ctx,
-					const unsigned char *abuf,
-					unsigned char *offset,
-					unsigned char *checksum,
-					const u64 Ls[16]) ASM_FUNC_ABI;
-#endif
-
 
 #ifdef USE_AMD64_ASM
 
@@ -841,9 +723,6 @@ extern void _gcry_twofish_amd64_decrypt_block(const TWOFISH_context *c,
 					      byte *out, const byte *in);
 
 /* These assembly implementations process three blocks in parallel. */
-extern void _gcry_twofish_amd64_blk3(const TWOFISH_context *c, byte *out,
-				     const byte *in, int encrypt);
-
 extern void _gcry_twofish_amd64_ctr_enc(const TWOFISH_context *c, byte *out,
 					const byte *in, byte *ctr);
 
@@ -865,58 +744,145 @@ extern void _gcry_twofish_amd64_ocb_auth(const TWOFISH_context *ctx,
 					 const byte *abuf, byte *offset,
 					 byte *checksum, const u64 Ls[3]);
 
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+static inline void
+call_sysv_fn (const void *fn, const void *arg1, const void *arg2,
+              const void *arg3, const void *arg4)
+{
+  /* Call SystemV ABI function without storing non-volatile XMM registers,
+   * as target function does not use vector instruction sets. */
+  asm volatile ("callq *%0\n\t"
+                : "+a" (fn),
+                  "+D" (arg1),
+                  "+S" (arg2),
+                  "+d" (arg3),
+                  "+c" (arg4)
+                :
+                : "cc", "memory", "r8", "r9", "r10", "r11");
+}
+
+static inline void
+call_sysv_fn5 (const void *fn, const void *arg1, const void *arg2,
+               const void *arg3, const void *arg4, const void *arg5)
+{
+  /* Call SystemV ABI function without storing non-volatile XMM registers,
+   * as target function does not use vector instruction sets. */
+  asm volatile ("movq %[arg5], %%r8\n\t"
+		"callq *%0\n\t"
+		: "+a" (fn),
+		  "+D" (arg1),
+		  "+S" (arg2),
+		  "+d" (arg3),
+		  "+c" (arg4)
+		: [arg5] "g" (arg5)
+		: "cc", "memory", "r8", "r9", "r10", "r11");
+}
+
+static inline void
+call_sysv_fn6 (const void *fn, const void *arg1, const void *arg2,
+               const void *arg3, const void *arg4, const void *arg5,
+	       const void *arg6)
+{
+  /* Call SystemV ABI function without storing non-volatile XMM registers,
+   * as target function does not use vector instruction sets. */
+  asm volatile ("movq %[arg5], %%r8\n\t"
+		"movq %[arg6], %%r9\n\t"
+		"callq *%0\n\t"
+		: "+a" (fn),
+		  "+D" (arg1),
+		  "+S" (arg2),
+		  "+d" (arg3),
+		  "+c" (arg4)
+		: [arg5] "g" (arg5),
+		  [arg6] "g" (arg6)
+		: "cc", "memory", "r8", "r9", "r10", "r11");
+}
+#endif
+
 static inline void
 twofish_amd64_encrypt_block(const TWOFISH_context *c, byte *out, const byte *in)
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn(_gcry_twofish_amd64_encrypt_block, c, out, in, NULL);
+#else
   _gcry_twofish_amd64_encrypt_block(c, out, in);
+#endif
 }
 
 static inline void
 twofish_amd64_decrypt_block(const TWOFISH_context *c, byte *out, const byte *in)
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn(_gcry_twofish_amd64_decrypt_block, c, out, in, NULL);
+#else
   _gcry_twofish_amd64_decrypt_block(c, out, in);
+#endif
 }
 
 static inline void
 twofish_amd64_ctr_enc(const TWOFISH_context *c, byte *out, const byte *in,
                       byte *ctr)
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn(_gcry_twofish_amd64_ctr_enc, c, out, in, ctr);
+#else
   _gcry_twofish_amd64_ctr_enc(c, out, in, ctr);
+#endif
 }
 
 static inline void
 twofish_amd64_cbc_dec(const TWOFISH_context *c, byte *out, const byte *in,
                       byte *iv)
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn(_gcry_twofish_amd64_cbc_dec, c, out, in, iv);
+#else
   _gcry_twofish_amd64_cbc_dec(c, out, in, iv);
+#endif
 }
 
 static inline void
 twofish_amd64_cfb_dec(const TWOFISH_context *c, byte *out, const byte *in,
                       byte *iv)
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn(_gcry_twofish_amd64_cfb_dec, c, out, in, iv);
+#else
   _gcry_twofish_amd64_cfb_dec(c, out, in, iv);
+#endif
 }
 
 static inline void
 twofish_amd64_ocb_enc(const TWOFISH_context *ctx, byte *out, const byte *in,
 		      byte *offset, byte *checksum, const u64 Ls[3])
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn6(_gcry_twofish_amd64_ocb_enc, ctx, out, in, offset, checksum, Ls);
+#else
   _gcry_twofish_amd64_ocb_enc(ctx, out, in, offset, checksum, Ls);
+#endif
 }
 
 static inline void
 twofish_amd64_ocb_dec(const TWOFISH_context *ctx, byte *out, const byte *in,
 		      byte *offset, byte *checksum, const u64 Ls[3])
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn6(_gcry_twofish_amd64_ocb_dec, ctx, out, in, offset, checksum, Ls);
+#else
   _gcry_twofish_amd64_ocb_dec(ctx, out, in, offset, checksum, Ls);
+#endif
 }
 
 static inline void
 twofish_amd64_ocb_auth(const TWOFISH_context *ctx, const byte *abuf,
 		       byte *offset, byte *checksum, const u64 Ls[3])
 {
+#ifdef HAVE_COMPATIBLE_GCC_WIN64_PLATFORM_AS
+  call_sysv_fn5(_gcry_twofish_amd64_ocb_auth, ctx, abuf, offset, checksum, Ls);
+#else
   _gcry_twofish_amd64_ocb_auth(ctx, abuf, offset, checksum, Ls);
+#endif
 }
 
 #elif defined(USE_ARM_ASM)
@@ -1129,7 +1095,7 @@ twofish_decrypt (void *context, byte *out, const byte *in)
 /* Bulk encryption of complete blocks in CTR mode.  This function is only
    intended for the bulk encryption feature of cipher.c.  CTR is expected to be
    of size TWOFISH_BLOCKSIZE. */
-static void
+void
 _gcry_twofish_ctr_enc(void *context, unsigned char *ctr, void *outbuf_arg,
 		      const void *inbuf_arg, size_t nblocks)
 {
@@ -1138,31 +1104,7 @@ _gcry_twofish_ctr_enc(void *context, unsigned char *ctr, void *outbuf_arg,
   const unsigned char *inbuf = inbuf_arg;
   unsigned char tmpbuf[TWOFISH_BLOCKSIZE];
   unsigned int burn, burn_stack_depth = 0;
-
-#ifdef USE_AVX2
-  if (ctx->use_avx2)
-    {
-      int did_use_avx2 = 0;
-
-      /* Process data in 16 block chunks. */
-      while (nblocks >= 16)
-        {
-          _gcry_twofish_avx2_ctr_enc(ctx, outbuf, inbuf, ctr);
-
-          nblocks -= 16;
-          outbuf += 16 * TWOFISH_BLOCKSIZE;
-          inbuf  += 16 * TWOFISH_BLOCKSIZE;
-          did_use_avx2 = 1;
-        }
-
-      if (did_use_avx2)
-        {
-          /* twofish-avx2 assembly code does not use stack */
-          if (nblocks == 0)
-            burn_stack_depth = 0;
-        }
-    }
-#endif
+  int i;
 
 #ifdef USE_AMD64_ASM
   {
@@ -1193,11 +1135,16 @@ _gcry_twofish_ctr_enc(void *context, unsigned char *ctr, void *outbuf_arg,
         burn_stack_depth = burn;
 
       /* XOR the input with the encrypted counter and store in output.  */
-      cipher_block_xor(outbuf, tmpbuf, inbuf, TWOFISH_BLOCKSIZE);
+      buf_xor(outbuf, tmpbuf, inbuf, TWOFISH_BLOCKSIZE);
       outbuf += TWOFISH_BLOCKSIZE;
       inbuf  += TWOFISH_BLOCKSIZE;
       /* Increment the counter.  */
-      cipher_block_add(ctr, 1, TWOFISH_BLOCKSIZE);
+      for (i = TWOFISH_BLOCKSIZE; i > 0; i--)
+        {
+          ctr[i-1]++;
+          if (ctr[i-1])
+            break;
+        }
     }
 
   wipememory(tmpbuf, sizeof(tmpbuf));
@@ -1207,7 +1154,7 @@ _gcry_twofish_ctr_enc(void *context, unsigned char *ctr, void *outbuf_arg,
 
 /* Bulk decryption of complete blocks in CBC mode.  This function is only
    intended for the bulk encryption feature of cipher.c. */
-static void
+void
 _gcry_twofish_cbc_dec(void *context, unsigned char *iv, void *outbuf_arg,
 		      const void *inbuf_arg, size_t nblocks)
 {
@@ -1217,31 +1164,6 @@ _gcry_twofish_cbc_dec(void *context, unsigned char *iv, void *outbuf_arg,
   unsigned char savebuf[TWOFISH_BLOCKSIZE];
   unsigned int burn, burn_stack_depth = 0;
 
-#ifdef USE_AVX2
-  if (ctx->use_avx2)
-    {
-      int did_use_avx2 = 0;
-
-      /* Process data in 16 block chunks. */
-      while (nblocks >= 16)
-        {
-          _gcry_twofish_avx2_cbc_dec(ctx, outbuf, inbuf, iv);
-
-          nblocks -= 16;
-          outbuf += 16 * TWOFISH_BLOCKSIZE;
-          inbuf  += 16 * TWOFISH_BLOCKSIZE;
-          did_use_avx2 = 1;
-        }
-
-      if (did_use_avx2)
-        {
-          /* twofish-avx2 assembly code does not use stack */
-          if (nblocks == 0)
-            burn_stack_depth = 0;
-        }
-    }
-#endif
-
 #ifdef USE_AMD64_ASM
   {
     /* Process data in 3 block chunks. */
@@ -1270,7 +1192,7 @@ _gcry_twofish_cbc_dec(void *context, unsigned char *iv, void *outbuf_arg,
       if (burn > burn_stack_depth)
         burn_stack_depth = burn;
 
-      cipher_block_xor_n_copy_2(outbuf, savebuf, iv, inbuf, TWOFISH_BLOCKSIZE);
+      buf_xor_n_copy_2(outbuf, savebuf, iv, inbuf, TWOFISH_BLOCKSIZE);
       inbuf += TWOFISH_BLOCKSIZE;
       outbuf += TWOFISH_BLOCKSIZE;
     }
@@ -1282,7 +1204,7 @@ _gcry_twofish_cbc_dec(void *context, unsigned char *iv, void *outbuf_arg,
 
 /* Bulk decryption of complete blocks in CFB mode.  This function is only
    intended for the bulk encryption feature of cipher.c. */
-static void
+void
 _gcry_twofish_cfb_dec(void *context, unsigned char *iv, void *outbuf_arg,
 		    const void *inbuf_arg, size_t nblocks)
 {
@@ -1291,31 +1213,6 @@ _gcry_twofish_cfb_dec(void *context, unsigned char *iv, void *outbuf_arg,
   const unsigned char *inbuf = inbuf_arg;
   unsigned int burn, burn_stack_depth = 0;
 
-#ifdef USE_AVX2
-  if (ctx->use_avx2)
-    {
-      int did_use_avx2 = 0;
-
-      /* Process data in 16 block chunks. */
-      while (nblocks >= 16)
-        {
-          _gcry_twofish_avx2_cfb_dec(ctx, outbuf, inbuf, iv);
-
-          nblocks -= 16;
-          outbuf += 16 * TWOFISH_BLOCKSIZE;
-          inbuf  += 16 * TWOFISH_BLOCKSIZE;
-          did_use_avx2 = 1;
-        }
-
-      if (did_use_avx2)
-        {
-          /* twofish-avx2 assembly code does not use stack */
-          if (nblocks == 0)
-            burn_stack_depth = 0;
-        }
-    }
-#endif
-
 #ifdef USE_AMD64_ASM
   {
     /* Process data in 3 block chunks. */
@@ -1342,7 +1239,7 @@ _gcry_twofish_cfb_dec(void *context, unsigned char *iv, void *outbuf_arg,
       if (burn > burn_stack_depth)
         burn_stack_depth = burn;
 
-      cipher_block_xor_n_copy(outbuf, iv, inbuf, TWOFISH_BLOCKSIZE);
+      buf_xor_n_copy(outbuf, iv, inbuf, TWOFISH_BLOCKSIZE);
       outbuf += TWOFISH_BLOCKSIZE;
       inbuf += TWOFISH_BLOCKSIZE;
     }
@@ -1351,7 +1248,7 @@ _gcry_twofish_cfb_dec(void *context, unsigned char *iv, void *outbuf_arg,
 }
 
 /* Bulk encryption/decryption of complete blocks in OCB mode. */
-static size_t
+size_t
 _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
 			const void *inbuf_arg, size_t nblocks, int encrypt)
 {
@@ -1359,49 +1256,10 @@ _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
   TWOFISH_context *ctx = (void *)&c->context.c;
   unsigned char *outbuf = outbuf_arg;
   const unsigned char *inbuf = inbuf_arg;
+  unsigned char l_tmp[TWOFISH_BLOCKSIZE];
   unsigned int burn, burn_stack_depth = 0;
   u64 blkn = c->u_mode.ocb.data_nblocks;
 
-#ifdef USE_AVX2
-  if (ctx->use_avx2)
-    {
-      int did_use_avx2 = 0;
-      u64 Ls[16];
-      u64 *l;
-
-      if (nblocks >= 16)
-	{
-          l = bulk_ocb_prepare_L_pointers_array_blk16 (c, Ls, blkn);
-
-	  /* Process data in 16 block chunks. */
-	  while (nblocks >= 16)
-	    {
-	      blkn += 16;
-	      *l = (uintptr_t)(void *)ocb_get_l(c, blkn - blkn % 16);
-
-	      if (encrypt)
-		_gcry_twofish_avx2_ocb_enc(ctx, outbuf, inbuf, c->u_iv.iv,
-					  c->u_ctr.ctr, Ls);
-	      else
-		_gcry_twofish_avx2_ocb_dec(ctx, outbuf, inbuf, c->u_iv.iv,
-					  c->u_ctr.ctr, Ls);
-
-	      nblocks -= 16;
-	      outbuf += 16 * TWOFISH_BLOCKSIZE;
-	      inbuf  += 16 * TWOFISH_BLOCKSIZE;
-	      did_use_avx2 = 1;
-	    }
-	}
-
-      if (did_use_avx2)
-	{
-	  /* twofish-avx2 assembly code does not use stack */
-	  if (nblocks == 0)
-	    burn_stack_depth = 0;
-	}
-    }
-#endif
-
   {
     /* Use u64 to store pointers for x32 support (assembly function
       * assumes 64-bit pointers). */
@@ -1410,9 +1268,10 @@ _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
     /* Process data in 3 block chunks. */
     while (nblocks >= 3)
       {
-	Ls[0] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 1);
-	Ls[1] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 2);
-	Ls[2] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 3);
+	/* l_tmp will be used only every 65536-th block. */
+	Ls[0] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 1);
+	Ls[1] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 2);
+	Ls[2] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 3);
 	blkn += 3;
 
 	if (encrypt)
@@ -1436,6 +1295,8 @@ _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
 
   c->u_mode.ocb.data_nblocks = blkn;
 
+  wipememory(&l_tmp, sizeof(l_tmp));
+
   if (burn_stack_depth)
     _gcry_burn_stack (burn_stack_depth + 4 * sizeof(void *));
 #else
@@ -1449,53 +1310,17 @@ _gcry_twofish_ocb_crypt (gcry_cipher_hd_t c, void *outbuf_arg,
 }
 
 /* Bulk authentication of complete blocks in OCB mode. */
-static size_t
+size_t
 _gcry_twofish_ocb_auth (gcry_cipher_hd_t c, const void *abuf_arg,
 			size_t nblocks)
 {
 #ifdef USE_AMD64_ASM
   TWOFISH_context *ctx = (void *)&c->context.c;
   const unsigned char *abuf = abuf_arg;
+  unsigned char l_tmp[TWOFISH_BLOCKSIZE];
   unsigned int burn, burn_stack_depth = 0;
   u64 blkn = c->u_mode.ocb.aad_nblocks;
 
-#ifdef USE_AVX2
-  if (ctx->use_avx2)
-    {
-      int did_use_avx2 = 0;
-      u64 Ls[16];
-      u64 *l;
-
-      if (nblocks >= 16)
-	{
-          l = bulk_ocb_prepare_L_pointers_array_blk16 (c, Ls, blkn);
-
-	  /* Process data in 16 block chunks. */
-	  while (nblocks >= 16)
-	    {
-	      blkn += 16;
-	      *l = (uintptr_t)(void *)ocb_get_l(c, blkn - blkn % 16);
-
-	      _gcry_twofish_avx2_ocb_auth(ctx, abuf, c->u_mode.ocb.aad_offset,
-					  c->u_mode.ocb.aad_sum, Ls);
-
-	      nblocks -= 16;
-	      abuf += 16 * TWOFISH_BLOCKSIZE;
-	      did_use_avx2 = 1;
-	    }
-	}
-
-      if (did_use_avx2)
-	{
-	  /* twofish-avx2 assembly code does not use stack */
-	  if (nblocks == 0)
-	    burn_stack_depth = 0;
-	}
-
-      /* Use generic code to handle smaller chunks... */
-    }
-#endif
-
   {
     /* Use u64 to store pointers for x32 support (assembly function
       * assumes 64-bit pointers). */
@@ -1504,13 +1329,14 @@ _gcry_twofish_ocb_auth (gcry_cipher_hd_t c, const void *abuf_arg,
     /* Process data in 3 block chunks. */
     while (nblocks >= 3)
       {
-	Ls[0] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 1);
-	Ls[1] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 2);
-	Ls[2] = (uintptr_t)(const void *)ocb_get_l(c, blkn + 3);
+	/* l_tmp will be used only every 65536-th block. */
+	Ls[0] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 1);
+	Ls[1] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 2);
+	Ls[2] = (uintptr_t)(const void *)ocb_get_l(c, l_tmp, blkn + 3);
 	blkn += 3;
 
 	twofish_amd64_ocb_auth(ctx, abuf, c->u_mode.ocb.aad_offset,
-			       c->u_mode.ocb.aad_sum, Ls);
+			      c->u_mode.ocb.aad_sum, Ls);
 
 	nblocks -= 3;
 	abuf += 3 * TWOFISH_BLOCKSIZE;
@@ -1525,6 +1351,8 @@ _gcry_twofish_ocb_auth (gcry_cipher_hd_t c, const void *abuf_arg,
 
   c->u_mode.ocb.aad_nblocks = blkn;
 
+  wipememory(&l_tmp, sizeof(l_tmp));
+
   if (burn_stack_depth)
     _gcry_burn_stack (burn_stack_depth + 4 * sizeof(void *));
 #else
@@ -1535,135 +1363,50 @@ _gcry_twofish_ocb_auth (gcry_cipher_hd_t c, const void *abuf_arg,
   return nblocks;
 }
 
+
 
-static unsigned int
-twofish_crypt_blk1_16(void *context, byte *out, const byte *in,
-		      size_t num_blks, int encrypt)
+/* Run the self-tests for TWOFISH-CTR, tests IV increment of bulk CTR
+   encryption.  Returns NULL on success. */
+static const char *
+selftest_ctr (void)
 {
-  TWOFISH_context *ctx = context;
-  unsigned int burn, burn_stack_depth = 0;
+  const int nblocks = 3+1;
+  const int blocksize = TWOFISH_BLOCKSIZE;
+  const int context_size = sizeof(TWOFISH_context);
 
-#ifdef USE_AVX2
-  if (num_blks == 16 && ctx->use_avx2)
-    {
-      _gcry_twofish_avx2_blk16 (ctx, out, in, encrypt);
-      return 0;
-    }
-#endif
-
-#ifdef USE_AMD64_ASM
-  while (num_blks >= 3)
-    {
-      _gcry_twofish_amd64_blk3 (ctx, out, in, encrypt);
-      burn = 8 * sizeof(void *);
-      burn_stack_depth = (burn > burn_stack_depth) ? burn : burn_stack_depth;
-      out += 3 * TWOFISH_BLOCKSIZE;
-      in += 3 * TWOFISH_BLOCKSIZE;
-      num_blks -= 3;
-    }
-#endif
-
-  while (num_blks >= 1)
-    {
-      if (encrypt)
-	burn = twofish_encrypt((void *)ctx, out, in);
-      else
-	burn = twofish_decrypt((void *)ctx, out, in);
-
-      burn_stack_depth = (burn > burn_stack_depth) ? burn : burn_stack_depth;
-      out += TWOFISH_BLOCKSIZE;
-      in += TWOFISH_BLOCKSIZE;
-      num_blks--;
-    }
-
-  return burn_stack_depth;
+  return _gcry_selftest_helper_ctr("TWOFISH", &twofish_setkey,
+           &twofish_encrypt, &_gcry_twofish_ctr_enc, nblocks, blocksize,
+	   context_size);
 }
 
-static unsigned int
-twofish_encrypt_blk1_16(void *ctx, byte *out, const byte *in,
-			size_t num_blks)
+/* Run the self-tests for TWOFISH-CBC, tests bulk CBC decryption.
+   Returns NULL on success. */
+static const char *
+selftest_cbc (void)
 {
-  return twofish_crypt_blk1_16 (ctx, out, in, num_blks, 1);
+  const int nblocks = 3+2;
+  const int blocksize = TWOFISH_BLOCKSIZE;
+  const int context_size = sizeof(TWOFISH_context);
+
+  return _gcry_selftest_helper_cbc("TWOFISH", &twofish_setkey,
+           &twofish_encrypt, &_gcry_twofish_cbc_dec, nblocks, blocksize,
+	   context_size);
 }
 
-static unsigned int
-twofish_decrypt_blk1_16(void *ctx, byte *out, const byte *in,
-			size_t num_blks)
+/* Run the self-tests for TWOFISH-CFB, tests bulk CBC decryption.
+   Returns NULL on success. */
+static const char *
+selftest_cfb (void)
 {
-  return twofish_crypt_blk1_16 (ctx, out, in, num_blks, 0);
+  const int nblocks = 3+2;
+  const int blocksize = TWOFISH_BLOCKSIZE;
+  const int context_size = sizeof(TWOFISH_context);
+
+  return _gcry_selftest_helper_cfb("TWOFISH", &twofish_setkey,
+           &twofish_encrypt, &_gcry_twofish_cfb_dec, nblocks, blocksize,
+	   context_size);
 }
 
-
-/* Bulk encryption/decryption of complete blocks in XTS mode. */
-static void
-_gcry_twofish_xts_crypt (void *context, unsigned char *tweak, void *outbuf_arg,
-			 const void *inbuf_arg, size_t nblocks, int encrypt)
-{
-  TWOFISH_context *ctx = context;
-  unsigned char *outbuf = outbuf_arg;
-  const unsigned char *inbuf = inbuf_arg;
-  int burn_stack_depth = 0;
-
-  /* Process remaining blocks. */
-  if (nblocks)
-    {
-      unsigned char tmpbuf[16 * 16];
-      unsigned int tmp_used = 16;
-      size_t tmpbufsize = 15 * 16;
-      size_t nburn;
-
-#ifdef USE_AVX2
-      if (ctx->use_avx2)
-	tmpbufsize = 16 * 16;
-#endif
-
-      nburn = bulk_xts_crypt_128(ctx, encrypt ? twofish_encrypt_blk1_16
-                                              : twofish_decrypt_blk1_16,
-                                 outbuf, inbuf, nblocks,
-                                 tweak, tmpbuf, tmpbufsize / 16,
-                                 &tmp_used);
-      burn_stack_depth = nburn > burn_stack_depth ? nburn : burn_stack_depth;
-
-      wipememory(tmpbuf, tmp_used);
-    }
-
-  if (burn_stack_depth)
-    _gcry_burn_stack(burn_stack_depth);
-}
-
-
-/* Bulk encryption/decryption in ECB mode. */
-static void
-_gcry_twofish_ecb_crypt (void *context, void *outbuf_arg, const void *inbuf_arg,
-			 size_t nblocks, int encrypt)
-{
-  TWOFISH_context *ctx = context;
-  unsigned char *outbuf = outbuf_arg;
-  const unsigned char *inbuf = inbuf_arg;
-  int burn_stack_depth = 0;
-
-  /* Process remaining blocks. */
-  if (nblocks)
-    {
-      size_t fn_maxblocks = 15;
-      size_t nburn;
-
-#ifdef USE_AVX2
-      if (ctx->use_avx2)
-	fn_maxblocks = 16;
-#endif
-
-      nburn = bulk_ecb_crypt_128(ctx, encrypt ? twofish_encrypt_blk1_16
-                                              : twofish_decrypt_blk1_16,
-                                 outbuf, inbuf, nblocks, fn_maxblocks);
-      burn_stack_depth = nburn > burn_stack_depth ? nburn : burn_stack_depth;
-    }
-
-  if (burn_stack_depth)
-    _gcry_burn_stack(burn_stack_depth);
-}
-
-
 
 /* Test a single encryption and decryption with each key size. */
 
@@ -1671,8 +1414,8 @@ static const char*
 selftest (void)
 {
   TWOFISH_context ctx; /* Expanded key. */
-  byte scratch[16];    /* Encryption/decryption result buffer. */
-  cipher_bulk_ops_t bulk_ops;
+  byte scratch[16];	/* Encryption/decryption result buffer. */
+  const char *r;
 
   /* Test vectors for single encryption/decryption.  Note that I am using
    * the vectors from the Twofish paper's "known answer test", I=3 for
@@ -1706,7 +1449,7 @@ selftest (void)
     0x05, 0x93, 0x1C, 0xB6, 0xD4, 0x08, 0xE7, 0xFA
   };
 
-  twofish_setkey (&ctx, key, sizeof(key), &bulk_ops);
+  twofish_setkey (&ctx, key, sizeof(key));
   twofish_encrypt (&ctx, scratch, plaintext);
   if (memcmp (scratch, ciphertext, sizeof (ciphertext)))
     return "Twofish-128 test encryption failed.";
@@ -1714,7 +1457,7 @@ selftest (void)
   if (memcmp (scratch, plaintext, sizeof (plaintext)))
     return "Twofish-128 test decryption failed.";
 
-  twofish_setkey (&ctx, key_256, sizeof(key_256), &bulk_ops);
+  twofish_setkey (&ctx, key_256, sizeof(key_256));
   twofish_encrypt (&ctx, scratch, plaintext_256);
   if (memcmp (scratch, ciphertext_256, sizeof (ciphertext_256)))
     return "Twofish-256 test encryption failed.";
@@ -1722,6 +1465,13 @@ selftest (void)
   if (memcmp (scratch, plaintext_256, sizeof (plaintext_256)))
     return "Twofish-256 test decryption failed.";
 
+  if ((r = selftest_ctr()) != NULL)
+    return r;
+  if ((r = selftest_cbc()) != NULL)
+    return r;
+  if ((r = selftest_cfb()) != NULL)
+    return r;
+
   return NULL;
 }
 
@@ -1742,8 +1492,7 @@ int
 main()
 {
   TWOFISH_context ctx;     /* Expanded key. */
-  int i, j;                /* Loop counters. */
-  cipher_bulk_ops_t bulk_ops;
+  int i, j;		    /* Loop counters. */
 
   const char *encrypt_msg; /* Message to print regarding encryption test;
                             * the printf is done outside the loop to avoid
@@ -1790,13 +1539,13 @@ main()
   /* Encryption test. */
   for (i = 0; i < 125; i++)
     {
-      twofish_setkey (&ctx, buffer[0], sizeof (buffer[0]), &bulk_ops);
+      twofish_setkey (&ctx, buffer[0], sizeof (buffer[0]));
       for (j = 0; j < 1000; j++)
         twofish_encrypt (&ctx, buffer[2], buffer[2]);
-      twofish_setkey (&ctx, buffer[1], sizeof (buffer[1]), &bulk_ops);
+      twofish_setkey (&ctx, buffer[1], sizeof (buffer[1]));
       for (j = 0; j < 1000; j++)
         twofish_encrypt (&ctx, buffer[3], buffer[3]);
-      twofish_setkey (&ctx, buffer[2], sizeof (buffer[2])*2, &bulk_ops);
+      twofish_setkey (&ctx, buffer[2], sizeof (buffer[2])*2);
       for (j = 0; j < 1000; j++) {
         twofish_encrypt (&ctx, buffer[0], buffer[0]);
         twofish_encrypt (&ctx, buffer[1], buffer[1]);
@@ -1808,15 +1557,15 @@ main()
   /* Decryption test. */
   for (i = 0; i < 125; i++)
     {
-      twofish_setkey (&ctx, buffer[2], sizeof (buffer[2])*2, &bulk_ops);
+      twofish_setkey (&ctx, buffer[2], sizeof (buffer[2])*2);
       for (j = 0; j < 1000; j++) {
         twofish_decrypt (&ctx, buffer[0], buffer[0]);
         twofish_decrypt (&ctx, buffer[1], buffer[1]);
       }
-      twofish_setkey (&ctx, buffer[1], sizeof (buffer[1]), &bulk_ops);
+      twofish_setkey (&ctx, buffer[1], sizeof (buffer[1]));
       for (j = 0; j < 1000; j++)
         twofish_decrypt (&ctx, buffer[3], buffer[3]);
-      twofish_setkey (&ctx, buffer[0], sizeof (buffer[0]), &bulk_ops);
+      twofish_setkey (&ctx, buffer[0], sizeof (buffer[0]));
       for (j = 0; j < 1000; j++)
         twofish_decrypt (&ctx, buffer[2], buffer[2]);
     }
